/**************************************************************
   This example shows how to set up a basic device and viewport
   ,set up a circle primitives and clut. It also shows how to
	spin them and change the graphic by adjusting the texture
	page offset.
**************************************************************/

#include "..\..\..\h\z_all.h"

#define		NAME	"Mr DD Demo 1"

#define		SCREEN_RES_X	320
#define		SCREEN_RES_Y	200


#define		MAX_SPHERES		128

int			app_active;
int			installed=-1;

/* object variables */
	VECTOR			object_trans={0,0,1024};
	SVECTOR			object_angle={0,0,0};
	MATRIX			object_matrix;

/* Device and viewport */
	DEVICE_TYPE		device;
	VIEWPORT_TYPE	view;

	char			*ramp_clut;

	CIRCLE			circles[MAX_SPHERES];

	SVECTOR			circle_centres[MAX_SPHERES];

	int				size=64;

void create_random_centres(SVECTOR *trans, int num_spheres)
{
	int		t;

	for (t=0; t<num_spheres; t++)
	{	
		/* Create a random translation in space */
		trans->vx=((rand()-16384) >> 3);
		trans->vy=((rand()-16384) >> 3);
		trans->vz=((rand()-16384) >> 3);

		trans++;
	}
}

void draw_spheres(SVECTOR *circle_trans)
{
	int		t, ot_pos;
	int		tpage;
	VECTOR	ov;
	long	p, flag;

	for (t=0; t<MAX_SPHERES; t++)
	{
		/* spin sphere using texture page start and change texture */
		tpage=((frame_count>>1)&127) + (t&3)*100*TPAGE_WIDTH;

		/* Rotate translate and add perspective to spheres centre */
		RotTrans(circle_trans, &ov, &flag);

		/* Scale order table position down a little so we don't drop off end */
		ot_pos=ov.vz>>2;

		/* Only draw if sphere is in front of viewer */
		if (ov.vz>VIEWPORT_current->z0)
		{
			/* set up circle structure */
			circles[t].tag=NULL;
			circles[t].id=MSPHERE|MTEXTURED;

			/* Setup spheres radius */
			circles[t].xr=128;
			circles[t].yr=128;

			/* Set up sheres 3D position */
			circles[t].cx=ov.vx;
			circles[t].cy=ov.vy;
			circles[t].cz=ov.vz;
			circles[t].r=FP_SHADE(15);

			/* Set up texture coardinates */
			circles[t].u[0]=0;
			circles[t].u[1]=FP_TEXT(127);
			circles[t].u[2]=FP_TEXT(127);
			circles[t].u[3]=0;
			circles[t].v[0]=0;
			circles[t].v[1]=0;
			circles[t].v[2]=FP_TEXT(100);
			circles[t].v[3]=FP_TEXT(100);
			circles[t].c=15;
			circles[t].tpage=tpage;

			/* Add primitive to viewport */
			VIEWPORT_add_prim(&view, (POLYGON *)&circles[t], ot_pos);
		}

		circle_trans++;
	}
}

int device_callback()
{
	frame_count++;

	return(NULL);
}


/*************************************************/
/* Main loop responsible for rotating models etc */
/*************************************************/
BOOL main_loop()
{
	char	err_str[16];
	int		err;

	object_angle.vx+=16;
	object_angle.vx&=4095;
	object_angle.vy+=16;
	object_angle.vy&=4095;

	/* Create rotation and translation matrix for object */
	RotMatrix(&object_angle, &object_matrix);
	TransMatrix(&object_matrix, &object_trans);

	/* Set as constant rotation and translation matrix */
	SetTransMatrix(&object_matrix);
	SetRotMatrix(&object_matrix);

	/* Draw the object */
	draw_spheres(&circle_centres[0]);

	/* Process the device and all viewports etc.. connected to it */
	err=DEVICE_process();
	if (err!=DEVICE_OK)
	{
		printf("DEVICE ERROR %d\n", err);

		return(FALSE);
	}
	frame_count++;

	return(TRUE);
}
BOOL setup_all()
{
	int					err;
	char				err_str[16];
	_RECT				view_dim;
	int					t;

	/* This is just so I can keep track of number of game frames */
	frame_count=0;

	/* Initialize zeus */
	if (SYS_init_zeus()==NULL)
	{
		printf("Could not initialize zeus\n");
		return(FALSE);
	}

	/* Request a clut (FLAT color model) */
	ramp_clut=SYS_request_clut_mem();
	if (ramp_clut==NULL)
	{
		printf("Not enough memory for ramp clut\n");
		return(FALSE);
	}
	
	tpage0=SYS_request_texture_page_mem(2*256);
	if (tpage0==NULL)
	{
		printf("Not enough memory for texture page\n");
		exit(1);
	}

	/* Clear device structure */
	DEVICE_clear(&device);

	/* Create primary device */
	err=DEVICE_create(&device, SCREEN_RES_X, SCREEN_RES_Y, 8, NULL,
			0, VIDEO_RAM_BASE_PC, DEVICE_AUTO_CLEAR, NULL, 1);
	if (err!=DEVICE_OK)
	{
		printf("DEVICE ERROR %d\n"), err;

		return(FALSE);
	}

	/* Turn vertical sync off */
	DEVICE_vertical_sync_off(&device);

	DEVICE_set_callback(&device, &device_callback);

	/* Load palette and set */
	DEVICE_load_palette(&device, "../newpal.pal");
	DEVICE_set_palette(&device);

	/* Create us a viewport */
	VIEWPORT_clear(&view);
	view_dim.x=0;
	view_dim.y=0;
	view_dim.w=SCREEN_RES_X-1;
	view_dim.h=SCREEN_RES_Y-1;
	err=VIEWPORT_create(&view, &view_dim, 0, FP7, 1024, DEVICE_NO_AUTO_CLEAR);
	if (err!=VIEWPORT_OK)
	{
		printf("VIEWPORT ERROR %d\n", err);

		return(FALSE);
	}
	VIEWPORT_set_border(&view, 15);			/* Tells viewport to display a border */
	VIEWPORT_set_min_size(&view, 32,8);		/* Minimized size */
	VIEWPORT_set_max_size(&view, 320,200);	/* Maximized size */
	VIEWPORT_set_near_z(&view, 0);			/* Font Z clipping plane value */
	VIEWPORT_set_interpolation(&view, 4);	/* interpolation during perspective correct texture mapped rendering */
	VIEWPORT_set_perspective(&view, 8);		/* Set eye distance from screen */

	/* Assign texture page to viewport */
	VIEWPORT_assign_tpage(&view, tpage0);

	/* Assign clut to viewport */
	VIEWPORT_assign_ramp_clut(&view, ramp_clut);

	/* Add the viewport to device */
	DEVICE_add_viewport(&device, &view);
				 
	/* Make 'device' the current working device */
	if (err=DEVICE_make_current(&device)<0)
	{
		printf("COULD NOT install devive \n");

		return(FALSE);
	}

	/* Make 'view' current working viewport */
	VIEWPORT_make_current(&view);

	/* Always call at least once before doing anything with it */
	DEVICE_process();

	/* Load texture raw date */
	FILE_load_tpage("../planets.bin", tpage0);

	/* Load ramp clut */
	FILE_load_clut("../pal_ref1.tab", ramp_clut);

	/* Create random sphere centres */
	create_random_centres(&circle_centres[0], MAX_SPHERES);

	return(TRUE);
}
void cleanup_all()
{
	/* Close down device, all viewports, order tables, primitive data, palettes etc.. */
	DEVICE_delete(&device);
}

void main()
{
	/* Make a device, viewport and load in graphics files etc */
	if (!(setup_all()))
	    return;

	installed=TRUE;

	/* Install keyboard handler */
	KBD_install();

	/* MAIN LOOP CALL */
	do
	{ 
		if (!main_loop())
			return;
	}
	while(!keybuf[KEY_ESC]);	/* continue until escape key pressed */

	/* kill keyboard handler */
	KBD_kill();

	cleanup_all();

	set_video_mode(3);			/* set text video mode */
}
