/*_________________________________________________________________________*\
                            \/ /              \ \/
                              <  Midi Example  >
                               \______________/

                          VARMINT'S AUDIO TOOLS 1.1


  This file is part of the VAT 1.1 distribution set. It is being provided
  as part of the VAT 1.1 demo distribution as a courtesy to prospective
  Ground Up customers. Unregistered users are granted a limited license to
  compile and run this code for demonstration purposes.  Please see the
  LICENSE.TXT file for information on how you may use the VAT library.

  This program demonstrates the VAT Midi API (Application Programming
  Interface).  Please refer to your VAT Reference Manual for information
  not covered here.  Correspondence should be directed to:

    support@groundup.com

  We very much encourage your comments and suggestions.  We want VAT to
  be useful and enjoyable for you.

    Please visit our web site at:

        http://www.groundup.com/

                                            Copyright 1995,1996 Ground Up
\*_________________________________________________________________________*/


#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
#include "vat.h"                     // VAT's header file, always mandatory

#define MIDI_FILE "sample4.mid"      // path to the Midi file example
#define CUSTOM_INST "custom.vcf"     // path to the instrument set example

  // a macro to simplify part of the example
#define neednewline() (tolower(lastcmd) == 'v' || tolower(lastcmd) == 't')

void report_chain (void);            // function for chain demo
volatile LONG chain_called;          // volatile marker for chain demo


//-------------------------------- main () --------------------------------//

int main ()
{
  MIDI *midi;       // MIDI structure pointer used with LoadMidi()
  char  err[128];   // error string used with LoadMidi()
  int   quit;       // quit flag used in play loop
  int   tempo;      // used to track tempo setting
  int   volume;     // used to track volume setting
  int   tmp;        // temporary variable
  char  c;          // used to hold input characters
  char  lastcmd;    // used in play loop to track last command for formatting
  INSTRUMENT *inst; // INSTRUMENT array pointer used with LoadInstruments()


  printf ("Initializing SoundBlaster at 45kHz.\n");
    //
    // Start VAT at 45Khz. As we are in a text mode the refresh setting
    // is not important; we'll use 70hz for it. If there is a problem,
    // SBSetUp() returns 0 and the sberr global variable is set to an
    // index into the errname global variable which can be used to report
    // the cause of the error.
    //
  if (!SBSetUp (45454, 70)) {
    printf ("Error in SBSetUp(): %s\n", errname[sberr]);
    exit (1);
  }
    //
    // SBSetUp() must be paired with GoVarmint() before any audio output.
    //
  GoVarmint ();
    //
    // Report the actual rate VAT was initialized at. This may be altered
    // from the rate specified above in two ways:
    //  1. The rate will be slightly adjusted to one of many predefined
    //      levels which were found to be optimal for audio output.
    //  2. If the VAT library used is the shareware or demo version, the
    //      rate will be limited to 11kHz.
    //
  printf ("Actually initialized to: %dHz\n", SampleRate ());


  printf ("Loading Midi '" MIDI_FILE "'...\n");
    //
    // Load the Midi. LoadMidi() allocates memory for the Midi it loads and
    // returns a pointer to it. Save that pointer in the midi variable.
    // VAT will not change the contents of the loaded Midi, so we'll use
    // the midi variable over and over in this example before we free it.
    //
  midi = LoadMidi (MIDI_FILE, err);
    //
    // LoadMidi() returns NULL on failure and writes the cause of failure
    // (up to 128 characters worth) to the passed error string. This is why
    // we have to pass a string which has already been allocated 128 bytes.
    //
  if (!midi) {
    printf ("\nError loading Midi: %s\n", err);
    exit (1);
  }
    //
    // Display the number of tracks in the loaded Midi. There are a few
    // interesting and informative statistics in each audio file type,
    // although most of the loaded data is obscure.
    //
  printf (" - %d tracks\n", midi->num_tracks);


  printf ("Loading custom instrument set '" CUSTOM_INST "'...\n");
    //
    // Load a custom instrument set. LoadInstruments() allocates memory for
    // the instrument set it loads and returns a pointer to it. Save that
    // pointer in the inst variable, and work with it as you would an array.
    // The included instrument set example is not very interesting, since
    // it is nearly identical to the default instrument set. You can use
    // Varmint's Voice Editor to create your own custom instrument sets.
    //
  inst = LoadInstruments (CUSTOM_INST);
    //
    // LoadInstruments() returns NULL on failure.
    //
  if (!inst) {
    printf ("Error loading custom instrument set.\n");
    exit (1);
  }


  printf ("\nPlaying Midi...\n");
    //
    // Start the Midi playing. This is the entry point for all Midi activity.
    //
  PlayMidi (midi);


    //
    // Wait a few milliseconds after the Midi starts playing for VAT to set
    // the Midi's tempo, then retrieve/report the volume and tempo settings.
    //
  MilliDelay (15);
  volume = MidiSetting (v_volume, GET_SETTING);
  tempo = MidiSetting (v_rate, GET_SETTING);
  printf ("<volume=%d, tempo=%d>\n", volume, tempo);


  printf ("Chaining Midi to function...\n");
    //
    // Use MidiChainFunc() to chain the local report_chain() function to
    // the current Midi. If there is a problem, a value of v_false will
    // be returned. v_false is an enumeration defined in vat.h whose
    // value is zero, so you could also simply check against zero. Also
    // initialize the local chain_called variable to zero, which tells
    // the play loop that the chain has not yet been called. Its value
    // will be set to nonzero when the chain is called.
    //
  if (!MidiChainFunc (report_chain))
    printf ("\nError chaining Midi.");
  chain_called = 0;


    //
    // Display information on how to control the Midi as it plays.
    //
  printf ("\nPlay Controls:\n"
          "  q  - quit\n"
          "  ?  - this listing\n"
          "  s  - start/stop\n"
          "  p  - toggle pause\n"
          "  r  - rewind\n"
          " V,v - raise,lower volume\n"
          " T,t - raise,lower tempo\n"
          "  i  - toggle custom instruments\n"
          "  f  - toggle FM output\n"
          "%s\n", MPUStatus () == v_unavail ?
                    "       (MPU output not available)" :
                    "  m  - toggle MPU output");

  printf ("<play>\n");    // note that playing has begun
  quit = 0;               // initialize the quit flag to zero


    //
    // Start the play loop, which will continue to run until the quit flag
    // is set to nonzero, which happens when the user hits the 'q' key.
    // Every time the user hits one of the valid command keys, the program
    // responds with confirmation of what was done. The way the volume and
    // tempo settings are managed and output makes the example a bit more
    // complex than it really needs to be. You can basically ignore calls
    // to printf() and fflush() and the macro neednewline() as part of the
    // formatting control.
    //
  while (!quit) {
      //
      // Check the local chain_called variable to see if it has been set
      // to a nonzero value, which happens in the report_chain() function,
      // which is called as a chained function when the Midi has finished
      // playing.
      //
    if (chain_called) {
      if (neednewline())
        printf ("\n");
      printf ("<finished playing, chain called, rechaining>\n");
        //
        // Since the chained function has been triggered, the Midi is no
        // longer chained. Use MidiChainFunc() again to rechain the
        // report_chain() function to the current Midi.
        //
      if (!MidiChainFunc (report_chain))
        printf ("Error chaining Midi.\n");
      chain_called = 0;
      lastcmd = 's';  // for the sake of formatting
    }

    if (!kbhit ())    // check for a keypress using the standard method
      continue;
    c = getch ();     // if a keypress was detected, grab it

      //
      // Scan through the valid command keys.
      //
    switch (c) {
    case 'Q':
    case 'q':
      if (neednewline())
        printf ("\n");
        //
        // Remove the current Midi with MidiCommand() and the v_remove
        // enumeration, resulting in there being no current Midi.
        //
      MidiCommand (v_remove);
      printf ("<remove>\n");
        //
        // Set the quit flag to nonzero, indicating the play loop should
        // be exited.
        //
      quit = 1;
      break;
    case '?':
      if (neednewline())
        printf ("\n");
        //
        // Display information on how to control the Midi as it plays.
        //
      printf ("\nPlay Controls:\n"
              "  q  - quit\n"
              "  ?  - this listing\n"
              "  s  - start/stop\n"
              "  p  - toggle pause\n"
              "  r  - rewind\n"
              " V,v - raise,lower volume\n"
              " T,t - raise,lower tempo\n"
              "  i  - toggle custom instruments\n"
              "  f  - toggle FM output\n"
              "%s\n", MPUStatus () == v_unavail ?
                        "       (MPU output not available)" :
                        "  m  - toggle MPU output");
      break;
    case 'S':
    case 's':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current Midi using MidiStatus(). If it is
        // stopped, restart it with MidiCommand() and the v_play enumeration.
        // Also re-initialize the tempo variable.
        //
      if (MidiStatus () == v_stopped) {
          //
          // Restart the Midi.
          //
        MidiCommand (v_play);
        printf ("<play>\n");
          //
          // Wait a few milliseconds after the Midi starts playing for VAT
          // to set the Midi's tempo and then retrieve that value.
          //
        MilliDelay (15);
        tmp = MidiSetting (v_rate, GET_SETTING);
        if (tmp != tempo) {
          tempo = tmp;
          printf ("<tempo reset to %d>\n", tempo);
        }
      } else { // the current Midi is not stopped
          //
          // The current Midi's status was reported to be something other
          // than v_stopped, so to toggle the started/stopped play state,
          // use MidiCommand() with the v_stop enumeration to stop the Midi.
          //
        MidiCommand (v_stop);
        printf ("<stop>\n");
      }
      break;
    case 'P':
    case 'p':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current Midi using MidiStatus(). If it is
        // paused, resume it with MidiCommand() and the v_play enumeration.
        // Otherwise, if it is playing, pause it with MidiCommand() and the
        // v_pause enumeration. Note that the second state check isn't
        // actually necessary; VAT will simply ignore a v_pause when the
        // Midi is either stopped or nonexistent. This is done here for the
        // sake of correctness and more state conscious reporting.
        //
      if (MidiStatus () == v_paused) {
        MidiCommand (v_play);
        printf ("<play>\n");
      } else if (MidiStatus () == v_started) {
        MidiCommand (v_pause);
        printf ("<pause>\n");
      }
      break;
    case 'R':
    case 'r':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current Midi using MidiStatus(). If it is
        // not stopped, rewind it with MidiCommand() and the v_rewind
        // enumeration. Note that it isn't actually necessary to check the
        // play state here; VAT will simply ignore a v_rewind when the Midi
        // is either stopped or nonexistent. This is done here for the sake
        // of correctness and more state conscious reporting.
        //
      if (MidiStatus () != v_stopped) {
        MidiCommand (v_rewind);
        printf ("<rewind>\n");
      }
      break;
    case 'V':
        //
        // Raise the volume setting by one using MidiSetting() with the
        // v_volume enumeration. Rather than making sure the value we're
        // setting is valid, we simply send the value and compare it against
        // the return value. If the return value is the same as the current
        // volume setting, we've reached the maximum volume setting.
        // Otherwise it worked and we assign the returned value to the
        // volume tracking variable.
        //
      tmp = MidiSetting (v_volume, volume+1);
      if (lastcmd == 'V')
        printf ("\r");
      else if (lastcmd == 'v' || tolower (lastcmd) == 't')
        printf ("\n");
      if (tmp == volume)
        printf ("<volume=%d> (max) \b", volume);
      else {
        volume = tmp;
        printf ("<volume=%d>       \b\b\b\b\b\b\b", volume);
      }
      fflush (stdout);
      break;
    case 'v':
        //
        // Lower the volume setting by one using the same methods as above.
        //
      tmp = MidiSetting (v_volume, volume-1);
      if (lastcmd == 'v')
        printf ("\r");
      else if (lastcmd == 'V' || tolower (lastcmd) == 't')
        printf ("\n");
      if (tmp == volume)
        printf ("<volume=%d> (min) \b", volume);
      else {
        volume = tmp;
        printf ("<volume=%d>       \b\b\b\b\b\b\b", volume);
      }
      fflush (stdout);
      break;
    case 'T':
        //
        // Lower the tempo setting (which actually causes the Midi to play
        // faster) by ten using the same methods as outlined above in the
        // volume adjusting section.
        //
      tmp = MidiSetting (v_rate, tempo-10);
      if (lastcmd == 'T')
        printf ("\r");
      else if (lastcmd == 't' || tolower (lastcmd) == 'v')
        printf ("\n");
      if (tmp == tempo)
        printf ("<tempo=%d> (max) \b", tempo);
      else {
        tempo = tmp;
        printf ("<tempo=%d>       \b\b\b\b\b\b\b", tempo);
      }
      fflush (stdout);
      break;
    case 't':
        //
        // Raise the tempo setting (which actually causes the Midi to play
        // faster) by ten using the same methods as outlined above in the
        // volume adjusting section.
        //
      tmp = MidiSetting (v_rate, tempo+10);
      if (lastcmd == 't')
        printf ("\r");
      else if (lastcmd == 'T' || tolower (lastcmd) == 'v')
        printf ("\n");
      if (tmp == tempo)
        printf ("<tempo=%d> (min) \b", tempo);
      else {
        tempo = tmp;
        printf ("<tempo=%d>       \b\b\b\b\b\b\b", tempo);
      }
      fflush (stdout);
      break;
    case 'I':
    case 'i':
      if (neednewline())
        printf ("\n");
        //
        // Check to see if VAT is using our custom instrument set for Midi
        // output using MidiInstruments() with the special GET_INSTRUMENTS
        // definition. Internally, VAT uses a pointer to determine which
        // instrument set to use rather than copying values. This results
        // in being able to check which instrument set VAT is using without
        // having to check all the instrument values, but it also means
        // that if you free an instrument set VAT is using, bad things will
        // happen as VAT tries to read instrument data from random data in
        // memory.
        //
      if (MidiInstruments (GET_INSTRUMENTS) == inst) {
          //
          // Pass NULL to MidiInstruments() to reset the instrument set
          // VAT uses for Midi output to the default instrument set.
          //
        MidiInstruments (NULL);
        printf ("<using default instrument set>\n");
      } else {
          //
          // Pass inst to MidiInstruments() to have VAT use our custom
          // instrument set for Midi output.
          //
        MidiInstruments (inst);
        printf ("<using custom instrument set>\n");
      }
      break;
    case 'F':
    case 'f':
      if (neednewline())
        printf ("\n");
        //
        // Toggle whether VAT is sending Midi output to the FM chip.
        // As explained in the Reference Manual, FMStatus() should never
        // return v_nonexist or v_unavail since a successful call to
        // FMEnable() is required for SBSetUp() to be successful.
        //
      if (FMStatus () != v_started) {
        FMEnable ();
        printf ("<FM output enabled>\n");
      } else {
        FMDisable ();
        printf ("<FM output disabled>\n");
      }
      break;
    case 'M':
    case 'm':
      if (neednewline())
        printf ("\n");
        //
        // Toggle whether VAT is sending Midi output to the MPU-401 port.
        // Since MPU output is not a basic element of VAT, we must check
        // the return value of MPUEnable() to make sure it's successful,
        // and it's possible for MPUStatus() to return v_unavail if the
        // MPU initialization fails. Also, if MPUStatus() will be v_nonexist
        // before the first call to MPUEnable().
        //
      switch (MPUStatus ()) {
      case v_nonexist:
        if (!MPUEnable ())
          printf ("<MPU output enable failed>\n");
        else
          printf ("<MPU output enabled> "
                  "(you may need to lower the volume to hear it)\n");
        break;
      case v_unavail:
        printf ("<MPU output not available>\n");
        break;
      case v_stopped:
        MPUEnable ();
        printf ("<MPU output enabled>\n");
        break;
      case v_started:
        MPUDisable ();
        printf ("<MPU output disabled>\n");
        break;
      }
      break;
    default:
      continue;  // skip the lastcmd assignment below
    }

    lastcmd = c;  // track the last command for formatting purposes
  }
  if (neednewline())
    printf ("\n");


    //
    // Free the space occupied by the loaded Midi.
    //
  FreeMidi (midi);


    //
    // Free the space occupied by the loaded custom instrument set.
    //
  free (inst);


    //
    // Shut down VAT and exit.
    //
  printf ("\nShutting down VAT.\n");
  DropDeadVarmint ();
  SBCleanUp ();
  return 0;
}


//---------------------------- report_chain () ----------------------------//

  //
  // This function is used in the function chaining demo. It is called by
  // VAT when the Midi it is chained to finishes playing.
  //
void report_chain ()
{
  chain_called = 1;
}

