/*_________________________________________________________________________*\
                            \/ /             \ \/
                              <  S3M Example  >
                               \_____________/

                          VARMINT'S AUDIO TOOLS 1.1


  This file is part of the VAT 1.1 distribution set. It is being provided
  as part of the VAT 1.1 demo distribution as a courtesy to prospective
  Ground Up customers. Unregistered users are granted a limited license to
  compile and run this code for demonstration purposes.  Please see the
  LICENSE.TXT file for information on how you may use the VAT library.

  This program demonstrates the VAT S3M API (Application Programming
  Interface).  Please refer to your VAT Reference Manual for information
  not covered here.  Correspondence should be directed to:

    support@groundup.com

  We very much encourage your comments and suggestions.  We want VAT to
  be useful and enjoyable for you.

    Please visit our web site at:

        http://www.groundup.com/

                                            Copyright 1995,1996 Ground Up
\*_________________________________________________________________________*/


#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
#include "vat.h"                   // VAT's header file, always mandatory

#define S3M_FILE "sample2.s3m"     // path to the Wave file example

  // a macro to simplify part of the example
#define neednewline() (tolower(lastcmd) == 'v' || tolower(lastcmd) == 't')

void report_chain (void);          // function for chain demo
volatile LONG chain_called;        // volatile marker for chain demo


//-------------------------------- main () --------------------------------//

int main ()
{
  S3M  *s3m;      // S3M structure pointer used with LoadS3M()
  char  err[128]; // error string used with LoadS3M()
  int   quit;     // quit flag used in play loop
  int   tempo;    // used to track tempo setting
  int   volume;   // used to track volume setting
  int   mask;     // used to track channel mask setting
  int   tmp;      // temporary variable
  char  c;        // used to hold input characters
  char  lastcmd;  // used in play loop to track last command for formatting


  printf ("Initializing SoundBlaster at 45kHz.\n");
    //
    // Start VAT at 45Khz. As we are in a text mode the refresh setting
    // is not important; we'll use 70hz for it. If there is a problem,
    // SBSetUp() returns 0 and the sberr global variable is set to an
    // index into the errname global variable which can be used to report
    // the cause of the error.
    //
  if (!SBSetUp (45454, 70)) {
    printf ("Error in SBSetUp(): %s\n", errname[sberr]);
    exit (1);
  }
    //
    // SBSetUp() must be paired with GoVarmint() before any audio output.
    //
  GoVarmint ();
    //
    // Report the actual rate VAT was initialized at. This may be altered
    // from the rate specified above in two ways:
    //  1. The rate will be slightly adjusted to one of many predefined
    //      levels which were found to be optimal for audio output.
    //  2. If the VAT library used is the shareware or demo version, the
    //      rate will be limited to 11kHz.
    //
  printf ("Actually initialized to: %dHz\n", SampleRate ());


  printf ("Loading S3M '" S3M_FILE "'...\n");
    //
    // Load the S3M. LoadS3M() allocates memory for the S3M it loads and
    // returns a pointer to it. Save that pointer in the s3m variable.
    // VAT will not change the contents of the loaded S3M, so we'll use
    // the s3m variable over and over in this example before we free it.
    //
  s3m = LoadS3M (S3M_FILE, err);
    //
    // LoadS3M() returns NULL on failure and writes the cause of failure
    // (up to 128 characters worth) to the passed error string. This is why
    // we have to pass a string which has already been allocated 128 bytes.
    //
  if (!s3m) {
    printf ("Error loading S3M: %s\n", err);
    exit (1);
  }
    //
    // Display some characteristics of the loaded S3M. There are a few
    // interesting and informative statistics in each audio file type,
    // although most of the loaded data is obscure.
    //
  printf (" (%s) - %d channels used\n\n", s3m->title, s3m->channelnum);


  printf ("\nPlaying S3M...\n");
    //
    // Start the S3M playing. This is the entry point for all S3M activity.
    //
  PlayS3M (s3m);


    //
    // Wait a few milliseconds after the S3M starts playing for VAT to set
    // the S3M's tempo, then retrieve/report the volume and tempo settings.
    //
  MilliDelay (15);
  volume = S3MSetting (v_volume, GET_SETTING);
  tempo = S3MSetting (v_rate, GET_SETTING);
  printf ("<volume=%d, tempo=%d>\n", volume, tempo);


  printf ("Chaining S3M to function...\n");
    //
    // Use S3MChainFunc() to chain the local report_chain() function to
    // the current S3M. If there is a problem, a value of v_false will
    // be returned. v_false is an enumeration defined in vat.h whose
    // value is zero, so you could also simply check against zero. Also
    // initialize the local chain_called variable to zero, which tells
    // the play loop that the chain has not yet been called. Its value
    // will be set to nonzero when the chain is called.
    //
  if (!S3MChainFunc (report_chain))
    printf ("Error chaining S3M.");
  chain_called = 0;


    //
    // Display information on how to control the S3M as it plays.
    //
  printf ("\nPlay Controls:\n"
          "  q  - quit\n"
          "  ?  - this listing\n"
          "  s  - start/stop\n"
          "  p  - toggle pause\n"
          "  r  - rewind\n"
          " V,v - raise,lower volume\n"
          " T,t - raise,lower tempo\n"
          " 1-8 - toggle 1-8 channel output\n"
          " !-* - toggle 9-16 channel output\n");

  printf ("<play>\n");    // note that playing has begun
  quit = 0;               // initialize the quit flag to zero


    //
    // Start the play loop, which will continue to run until the quit flag
    // is set to nonzero, which happens when the user hits the 'q' key.
    // Every time the user hits one of the valid command keys, the program
    // responds with confirmation of what was done. The way the volume and
    // tempo settings are managed and output makes the example a bit more
    // complex than it really needs to be. You can basically ignore calls
    // to printf() and fflush() and the macro neednewline() as part of the
    // formatting control.
    //
  while (!quit) {
      //
      // Check the local chain_called variable to see if it has been set
      // to a nonzero value, which happens in the report_chain() function,
      // which is called as a chained function when the S3M has finished
      // playing.
      //
    if (chain_called) {
      if (neednewline())
        printf ("\n");
      printf ("<finished playing, chain called, rechaining>\n");
        //
        // Since the chained function has been triggered, the S3M is no
        // longer chained. Use S3MChainFunc() again to rechain the
        // report_chain() function to the current S3M.
        //
      if (!S3MChainFunc (report_chain))
        printf ("Error chaining S3M.\n");
      chain_called = 0;
      lastcmd = 's';  // for the sake of formatting
    }

    if (!kbhit ())    // check for a keypress using the standard method
      continue;
    c = getch ();     // if a keypress was detected, grab it

      //
      // Scan through the valid command keys.
      //
    switch (c) {
    case 'Q':
    case 'q':
      if (neednewline())
        printf ("\n");
        //
        // Remove the current S3M with S3MCommand() and the v_remove
        // enumeration, resulting in there being no current S3M.
        //
      S3MCommand (v_remove);
      printf ("<remove>\n");
        //
        // Set the quit flag to nonzero, indicating the play loop should
        // be exited.
        //
      quit = 1;
      break;
    case '?':
      if (neednewline())
        printf ("\n");
        //
        // Display information on how to control the S3M as it plays.
        //
      printf ("\nPlay Controls:\n"
              "  q  - quit\n"
              "  ?  - this listing\n"
              "  s  - start/stop\n"
              "  p  - toggle pause\n"
              "  r  - rewind\n"
              " V,v - raise,lower volume\n"
              " T,t - raise,lower tempo\n"
              " 1-8 - toggle 1-8 channel output\n"
              " !-* - toggle 9-16 channel output\n");
      break;
    case 'S':
    case 's':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current S3M using S3MStatus(). If it is
        // stopped, restart it with S3MCommand() and the v_play enumeration.
        // Also re-initialize the tempo variable and the channel mask.
        //
      if (S3MStatus () == v_stopped) {
          //
          // Restart the S3M.
          //
        S3MCommand (v_play);
        printf ("<play>\n");
          //
          // Wait a few milliseconds after the S3M starts playing for VAT
          // to set the S3M's tempo and then retrieve that value.
          //
        MilliDelay (15);
        tmp = S3MSetting (v_rate, GET_SETTING);
        if (tmp != tempo) {
          tempo = tmp;
          printf ("<tempo reset to %d>\n", tempo);
        }
          //
          // Reset the channel mask. The mask check here isn't actually
          // necessary; we're just resetting the value anyway. This is
          // done here for the sake of correctness and more state conscious
          // reporting.
          //
        if (S3MChannelMask (GET_SETTING)) {
          S3MChannelMask (0);
          printf ("<all channels reactivated>\n");
        }
      } else { // the current S3M is not stopped
          //
          // The current S3M's status was reported to be something other
          // than v_stopped, so to toggle the started/stopped play state,
          // use S3MCommand() with the v_stop enumeration to stop the S3M.
          //
        S3MCommand (v_stop);
        printf ("<stop>\n");
      }
      break;
    case 'P':
    case 'p':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current S3M using S3MStatus(). If it is
        // paused, resume it with S3MCommand() and the v_play enumeration.
        // Otherwise, if it is playing, pause it with S3MCommand() and the
        // v_pause enumeration. Note that the second state check isn't
        // actually necessary; VAT will simply ignore a v_pause when the
        // S3M is either stopped or nonexistent. This is done here for the
        // sake of correctness and more state conscious reporting.
        //
      if (S3MStatus () == v_paused) {
        S3MCommand (v_play);
        printf ("<play>\n");
      } else if (S3MStatus () == v_started) {
        S3MCommand (v_pause);
        printf ("<pause>\n");
      }
      break;
    case 'R':
    case 'r':
      if (neednewline())
        printf ("\n");
        //
        // Check the status of the current S3M using S3MStatus(). If it is
        // not stopped, rewind it with S3MCommand() and the v_rewind
        // enumeration. Note that it isn't actually necessary to check the
        // play state here; VAT will simply ignore a v_rewind when the S3M
        // is either stopped or nonexistent. This is done here for the sake
        // of correctness and more state conscious reporting.
        //
      if (S3MStatus () != v_stopped) {
        S3MCommand (v_rewind);
        printf ("<rewind>\n");
      }
      break;
    case 'V':
        //
        // Raise the volume setting by one using S3MSetting() with the
        // v_volume enumeration. Rather than making sure the value we're
        // setting is valid, we simply send the value and compare it against
        // the return value. If the return value is the same as the current
        // volume setting, we've reached the maximum volume setting.
        // Otherwise it worked and we assign the returned value to the
        // volume tracking variable.
        //
      tmp = S3MSetting (v_volume, volume+1);
      if (lastcmd == 'V')
        printf ("\r");
      else if (lastcmd == 'v' || tolower (lastcmd) == 't')
        printf ("\n");
      if (tmp == volume)
        printf ("<volume=%d> (max) \b", volume);
      else {
        volume = tmp;
        printf ("<volume=%d>       \b\b\b\b\b\b\b", volume);
      }
      fflush (stdout);
      break;
    case 'v':
        //
        // Lower the volume setting by one using the same methods as above.
        //
      tmp = S3MSetting (v_volume, volume-1);
      if (lastcmd == 'v')
        printf ("\r");
      else if (lastcmd == 'V' || tolower (lastcmd) == 't')
        printf ("\n");
      if (tmp == volume)
        printf ("<volume=%d> (min) \b", volume);
      else {
        volume = tmp;
        printf ("<volume=%d>       \b\b\b\b\b\b\b", volume);
      }
      fflush (stdout);
      break;
    case 'T':
        //
        // Lower the tempo setting (which actually causes the S3M to play
        // faster) by ten using the same methods as outlined above in the
        // volume adjusting section.
        //
      tmp = S3MSetting (v_rate, tempo-10);
      if (lastcmd == 'T')
        printf ("\r");
      else if (lastcmd == 't' || tolower (lastcmd) == 'v')
        printf ("\n");
      if (tmp == tempo)
        printf ("<tempo=%d> (max) \b", tempo);
      else {
        tempo = tmp;
        printf ("<tempo=%d>       \b\b\b\b\b\b\b", tempo);
      }
      fflush (stdout);
      break;
    case 't':
        //
        // Raise the tempo setting (which actually causes the S3M to play
        // faster) by ten using the same methods as outlined above in the
        // volume adjusting section.
        //
      tmp = S3MSetting (v_rate, tempo+10);
      if (lastcmd == 't')
        printf ("\r");
      else if (lastcmd == 'T' || tolower (lastcmd) == 'v')
        printf ("\n");
      if (tmp == tempo)
        printf ("<tempo=%d> (min) \b", tempo);
      else {
        tempo = tmp;
        printf ("<tempo=%d>       \b\b\b\b\b\b\b", tempo);
      }
      fflush (stdout);
      break;
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '!':
    case '@':
    case '#':
    case '$':
    case '%':
    case '^':
    case '&':
    case '*':
      if (c == '!')
        tmp = 8;
      else if (c == '@')
        tmp = 9;
      else if (c == '#')
        tmp = 10;
      else if (c == '$')
        tmp = 11;
      else if (c == '%')
        tmp = 12;
      else if (c == '^')
        tmp = 13;
      else if (c == '&')
        tmp = 14;
      else if (c == '*')
        tmp = 15;
      else
        tmp = c - '0' - 1;
        //
        // Toggle output of the indicated channel with S3MChannelMask().
        // Note the numbering begins at 1 for user accessibility but the
        // channel numbers are 0-15. This is accomplished by setting bits
        // in the mask value corresponding with channels we want to mute.
        // This program doesn't need to keep track of the mask setting
        // with the help of the special GET_SETTING definition, available
        // for nearly all value setting functions in VAT.
        //
      mask = S3MChannelMask (GET_SETTING);  // get the current channel mask
      if (neednewline())
        printf ("\n");
      if (mask & (1<<tmp)) {                // if the bit is set
        mask &= ~(1<<tmp);                  // clear it
        printf ("<ch%d active>\n", tmp+1);
      } else {
        mask |= (1<<tmp);                   // otherwise
        printf ("<ch%d muted>\n", tmp+1);   // set it
      }
      S3MChannelMask (mask);                // and send the new mask
      break;
    default:
      continue;  // skip the lastcmd assignment below
    }

    lastcmd = c;  // track the last command for formatting purposes
  }
  if (neednewline())
    printf ("\n");


    //
    // Free the space occupied by the loaded S3M.
    //
  FreeS3M (s3m);


    //
    // Shut down VAT and exit.
    //
  printf ("\nShutting down VAT.\n");
  DropDeadVarmint ();
  SBCleanUp ();
  return 0;
}


//---------------------------- report_chain () ----------------------------//

  //
  // This function is used in the function chaining demo. It is called by
  // VAT when the S3M it is chained to finishes playing.
  //
void report_chain ()
{
  chain_called = 1;
}

