/*_________________________________________________________________________*\
                           \/ /              \ \/
                             <  Wave Example  >
                              \______________/

                          VARMINT'S AUDIO TOOLS 1.1d


  This file is part of the VAT 1.1 distribution set. It is being provided
  as part of the VAT 1.1 demo distribution as a courtesy to prospective
  Ground Up customers. Unregistered users are granted a limited license to
  compile and run this code for demonstration purposes.  Please see the
  LICENSE.TXT file for information on how you may use the VAT library.

  This program demonstrates the VAT Wave API (Application Programming
  Interface).  Please refer to your VAT Reference Manual for information
  not covered here.  Correspondence should be directed to:

    support@groundup.com

  We very much encourage your comments and suggestions.  We want VAT to
  be useful and enjoyable for you.

    Please visit our web site at:

        http://www.groundup.com/

                                            Copyright 1995,1996 Ground Up
\*_________________________________________________________________________*/


#include <stdio.h>
#include <stdlib.h>
#include "vat.h"                // VAT's header file, always mandatory

#define WAVE_FILE "aha.wav"     // path to the Wave file example

#define DOWN 0
#define UP   1                  // some definitions to clarify the examples
#define LEFT  0
#define RIGHT 1

void report_chain (LONG);       // function for chain demo
volatile LONG chain_called=0;   // volatile marker for chain demo


//-------------------------------- main () --------------------------------//

int main ()
{
  WAVE *wav;      // WAVE structure pointer used with LoadWave()
  char  err[128]; // error string used with LoadWave()
  LONG  wave_id;  // a Wave handle
  LONG  chain_id; // a second Wave handle for chain demo
  int   freq;     // used in frequency demo
  int   pan;      // used in panning demo
  int   dir;      // direction flag used with frequency and panning demo


  printf ("Initializing SoundBlaster at 11kHz.\n");
    //
    // Start VAT at 11Khz. As we are in a text mode the refresh setting
    // is not important; we'll use 70hz for it. If there is a problem,
    // SBSetUp() returns 0 and the sberr global variable is set to an
    // index into the errname global variable which can be used to report
    // the cause of the error.
    //
  if (!SBSetUp (11000, 70)) {
    printf ("Error in SBSetUp(): %s\n", errname[sberr]);
    exit (1);
  }
    //
    // SBSetUp() must be paired with GoVarmint() before any audio output.
    //
  GoVarmint ();
    //
    // Report the actual rate VAT was initialized at. This may be altered
    // from the rate specified above in two ways:
    //  1. The rate will be slightly adjusted to one of many predefined
    //      levels which were found to be optimal for audio output.
    //  2. If the VAT library used is the shareware or demo version, the
    //      rate will be limited to 11kHz.
    //
  printf ("Actually initialized to: %dHz\n", SampleRate ());


  printf ("Loading Wave '" WAVE_FILE "'...\n");
    //
    // Load the Wave. LoadWave() allocates memory for the Wave it loads
    // and returns a pointer to it. Save that pointer in the wav variable.
    // VAT will not change the contents of the loaded Wave, so we'll use
    // the wav variable over and over in this example before we free it.
    //
  wav = LoadWave (WAVE_FILE, err);
    //
    // LoadWave() returns NULL on failure and writes the cause of failure
    // (up to 128 characters worth) to the passed error string. This is why
    // we have to pass a string which has already been allocated 128 bytes.
    //
  if (!wav) {
    printf ("Error loading Wave: %s\n", err);
    exit (1);
  }
    //
    // Display some characteristics of the loaded Wave. There are a few
    // interesting and informative statistics in each audio file type,
    // although most of the loaded data is obscure.
    //
  printf (" (recorded at %d Hz in %s)\n", wav->sample_rate,
          wav->stereo ? "stereo" : "mono");


    //
    // Play the Wave using the 'plain' play type. Because the plain type
    // allows no altering of frequency, if the Wave was recorded at a
    // different rate than VAT was initialized with, the Wave will sound
    // faster or slower than how it was recorded.
    //
  printf ("\nPlaying as type v_plain...\n");
    //
    // Keep track of the Wave using the handle PlayWave() returns. Handles
    // are only used with Waves, since multiple Waves may be played at once,
    // and are used to check play status and modify play settings.
    //
  wave_id = PlayWave (wav, 0, v_plain);
    //
    // If there's a problem, the return value will be 0.
    //
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else
      //
      // When a Wave finishes playing, it is removed from the play queue,
      // altering its status to 'nonexistent'. Keep checking for this
      // condition.
      //
    while (WaveStatus (wave_id) != v_nonexist);


    //
    // Shut down VAT.
    //
  printf ("\nShutting down VAT.\n");
  DropDeadVarmint ();
  SBCleanUp ();


    //
    // Restart VAT at 45Khz, and as before we'll just use 70hz for the
    // refresh rate.
    //
  printf ("\nRe-initializing SoundBlaster at 45kHz.\n");
  fflush (stdout);
  if (!SBSetUp (45000, 70)) {   // 45Khz at 70hz refresh
    printf ("Error in SBSetUp(): %s\n", errname[sberr]);
    exit (1);
  }
  GoVarmint ();
  printf ("Actually initialized to: %dHz\n", SampleRate ());


    //
    // Play the Wave using the 'plain' play type again. Once again, since
    // the plain play type allows no altering of frequency, if the Wave was
    // recorded at a different rate than VAT was initialized with, the Wave
    // will sound faster or slower than how it was recorded. Since VAT has
    // been initialized this time at a different rate (45kHz) than it was
    // last time we played the Wave (11kHz), it will definitely sound
    // different than the last time we played it.
    //
  printf ("\nPlaying as type v_plain...\n");
  wave_id = PlayWave (wav, 0, v_plain);          // play as v_plain
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else
    while (WaveStatus (wave_id) != v_nonexist);  // wait for it to finish


    //
    // Play the Wave using the 'fancy' play type. This play type does allow
    // for altering of frequency, and in fact VAT will automatically set the
    // play frequency to the rate the Wave was recorded at (stored in the
    // WAVE structure). This means the Wave should sound the same as how it
    // was recorded.
    //
  printf ("Playing as type v_fancy...\n");
  wave_id = PlayWave (wav, 0, v_plain);          // play as v_fancy
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else
    while (WaveStatus (wave_id) != v_nonexist);  // wait for it to finish


    //
    // Play the Wave using the 'fancy' play type and then tweak its frequency
    // up and down on the fly while we're waiting for it to finish playing.
    //
  printf ("Playing as type v_fancy and altering frequency...\n");
  wave_id = PlayWave (wav, 0, v_fancy);
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else {
    freq = wav->sample_rate; // init 'freq' to recorded frequency
    dir = DOWN;              // init 'dir' to start moving frequency down

    while (WaveStatus (wave_id) != v_nonexist) { // while it hasn't finished
      if (dir == UP) {
        freq += 1000;
        if (freq + 1000 > 45454)                 // tweak the frequency up
          dir = DOWN;                            // until we pass 45kHz
      } else {
        freq -= 1000;
        if (freq - 1000 < 4000)                  // and tweak it down
          dir = UP;                              // until we pass 4kHz
      }
        //
        // Use WaveSetting() with the saved PlayWave() handle to set the
        // frequency to the new value. In this case we'll ignore the return
        // value, although it would inform us of the outcome of the setting,
        // including a failure for invalid setting type or any adjustments
        // made to the value.
        //
      WaveSetting (wave_id, v_rate, freq);
        //
        // Wait 50 milliseconds before tweaking the frequency again.
        //
      MilliDelay (50);
    }
  }


    //
    // Play the Wave using the 'pan' play type and then tweak its pan position
    // back and forth on the fly while we're waiting for it to finish playing.
    //
  printf ("Playing as type v_pan and altering pan position...\n");
  wave_id = PlayWave (wav, 0, v_pan);
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else {
    pan = 30;   // init 'pan' to middle
    dir = LEFT; // init 'dir' to start moving pan position to the left

    while (WaveStatus (wave_id) != v_nonexist) { // while it hasn't finished
      if (dir == RIGHT) {
        pan++;
        if (pan >= 59)                           // tweak the pan pos. right
          dir = LEFT;                            // until we pass 59
      } else {
        pan--;
        if (pan <= 1)                            // and tweak it left
          dir = RIGHT;                           // until we pass 1
      }
        //
        // Use WaveSetting() with the saved PlayWave() handle to set the
        // pan position to the new value. Again, as above, we'll ignore the
        // return value.
        //
      WaveSetting (wave_id, v_panpos, pan);
        //
        // This time we'll wait 10 milliseconds between panning tweaks.
        //
      MilliDelay (10);
    }
  }


    //
    // Play the Wave using the 'fancy' play type and then chain it to
    // itself. We're only using v_fancy so it will sound correct; that's
    // not necessary for a chain to work.
    //
  printf ("Playing as type v_fancy and chaining to itself...\n");
  wave_id = PlayWave (wav, 0, v_fancy);
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else {
      //
      // Use WaveChain() with the saved PlayWave() handle to chain a new
      // copy of the Wave to the already-queued Wave. Keep track of the new
      // chained Wave using the handle WaveChain() returns. This handle
      // won't really be useful until the first Wave finishes playing and
      // this new Wave is started, and until then the chained Wave's status
      // will be v_stopped.
      //
    chain_id = WaveChain (wave_id, wav, v_fancy);
      // 
      // If there is a problem, the return value will be 0, just like with
      // PlayWave().
      //
    if (!chain_id)
      printf ("Error chaining Wave to Wave.\n");
    else {
        //
        // Wait for the first Wave to finish playing by watching wave_id.
        //
      while (WaveStatus (wave_id) != v_nonexist);
        //
        // At this point, the chained Wave should kick in.
        //
      printf ("  Chained sound playing...\n");
        //
        // Wait for the chained Wave to finish playing by watching chain_id.
        //
      while (WaveStatus (chain_id) != v_nonexist);
    }
  }


    //
    // Play the Wave using the 'fancy' play type and then chain a function
    // to it. As above, we're only using v_fancy so it will sound correct;
    // you can chain with any play type.
    //
  printf ("Playing as type v_fancy and chaining to a function...\n");
  wave_id = PlayWave (wav, 0, v_fancy);
  if (!wave_id)
    printf ("Error playing Wave.\n");
  else {
      //
      // As with WaveChain() above, use WaveChainFunc() with the saved
      // PlayWave() handle to chain the local report_chain() function to
      // the Wave.
      //
    if (!WaveChainFunc (wave_id, report_chain))
      printf ("Error chaining Wave to function.\n");
    else {
        //
        // Wait for the first Wave to finish playing by watching wave_id.
        //
      while (WaveStatus (wave_id) != v_nonexist);
        //
        // At this point, the chained function should kick in.
        //
      printf ("  Chained function running...\n");
        //
        // The local chain_called variable is initialized to zero and
        // set to a nonzero value when the chained function report_chain()
        // runs. Keep checking for this condition.
        //
      while (!chain_called)
        MilliDelay (10);
        //
        // Print some verification information. A chained function will
        // be passed the handle of the terminating Wave when VAT calls it.
        //
      printf ("  (wave_id: %d, chain called with: %d)\n",
              wave_id, chain_called);
    }
  }


    //
    // Free the space occupied by the loaded Wave.
    //
  FreeWave (wav);


    //
    // Shut down VAT and exit.
    //
  printf ("\nShutting down VAT.\n");
  DropDeadVarmint ();
  SBCleanUp ();
  return(0);
}


//---------------------------- report_chain () ----------------------------//

  //
  // This function is used in the function chaining demo about 50 lines up.
  // It is called by VAT when the Wave it is chained to finishes playing,
  // and gets passed the handle of that Wave, which it assigns to the local
  // volatile chain_called variable.
  //
void report_chain (LONG chain_id)
{
  chain_called = chain_id;
}

